/****************************************************  Utility Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to the Signal Stream application example.

	Rev History: 1-10-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

****************************************************  Utility Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "adapter.h"
#include "sigfpga.h"
#include "common_functions.h"
#include "channel_functions.h"
#include "dma_functions.h"

// Write the contents of the requested number of DMA pages associated with the requested channel number to the specified file.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long DataSave(void **p_PageAddresses, s_Channel RXChannel, char *p_File, size_t PageCount, size_t DataItemSize)
{
	unsigned long Status = 0;
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	size_t page_count = 0;
	size_t data_items_count = 0;
	size_t record_count = 0;
	size_t records_per_page = 0;
	signed sample = 0;
	unsigned long long *virtual_address;

	// Open the specified data file.
	fp = fopen(p_File, "w");

	// Process samples only if the file open was successful.
	if(fp != NULL)
	{
		// Compute the number of 64-bit records per DMA page.
		records_per_page = RXChannel.DMABuffer.BurstSizeB * RXChannel.DMABuffer.BurstCount / 8;

		// Sequence through all of the DMA pages assigned to the channel.
		for(page_count=0; page_count < PageCount; page_count++)
		{
			// Get the virtual address of the first 64-bit record in each page.
			virtual_address = (unsigned long long *)p_PageAddresses[page_count];

			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < records_per_page; record_count++)

				// Sequence through every sample in the 64-bit record.
				for (data_items_count = 0; data_items_count < (64 / DataItemSize); data_items_count++)
					{
						// Extract the next sample value from the 64-bit record.
						sample = (signed)(*(virtual_address + record_count) >> (data_items_count * DataItemSize)) & (max >> (64 - DataItemSize));
						// Sign extend negative numbers.
						if (sample & (one << (DataItemSize - 1))) sample |= (max << DataItemSize);
						// Print the sample value to the file.
						fprintf(fp, "%d\n", sample);
					}
		}
		fclose(fp);
		Status = 0;
	}
	else
		Status = 1;

	return(Status);
}

// Preload a transmitter buffer with sine wave data samples.
// Return of zero indicates that the buffer contains an integer number of sine wave periods, non-zero indicates a fractional number of periods.
unsigned long DataLoad(void **p_PageAddresses, s_Channel TXChannel, double FsDivisor, size_t DataItemSize)
{
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	size_t page_count = 0;
	size_t record_count = 0;
	size_t records_per_page = 0;
	size_t data_items_per_record = 0;
	size_t data_items_count = 0;
	unsigned long long *virtual_address;
	double phase = 0;
	double phase_select = 0;
	double data_item_size_dbl = 0;

	// Cast the DataItemSize variable for use with other data types.
	data_item_size_dbl = (double)DataItemSize; 

	// Compute the number of 64-bit records per DMA page.
	records_per_page = TXChannel.DMABuffer.BurstSizeB * TXChannel.DMABuffer.BurstCount / 8;

	// Compute the number of data items per 64-bit record.
	data_items_per_record = (size_t)floor(64 / data_item_size_dbl);

	// Sequence through the number of DMA pages per mark since this one chapter will be used for every DMA buffer refresh.
	for(page_count=0; page_count < TXChannel.DMABuffer.PagesPerMark; page_count++)
	{
		// Get the virtual address of the first 64-bit record in the current page.
		virtual_address = (unsigned long long *)p_PageAddresses[page_count];

		// Sequence through every 64-bit record in the page.
		for(record_count = 0; record_count < records_per_page; record_count++)
		{
			// Clear the current 64-bit record.
			*(virtual_address + record_count) = 0x0ull;
			
			for(data_items_count = 1; data_items_count <= data_items_per_record; data_items_count++)
			{
				// Compute the amplitude of the DAC sample based on phase and write it as the current data item to the payload record.
				*(virtual_address + record_count) |= (((max >> (64-DataItemSize)) & (unsigned long long)((pow(2,data_item_size_dbl-1)-1)*sin(2*3.141592654*phase/FsDivisor))) << (64-(data_items_count*DataItemSize)));

				// Compute the phase of the next sample.
				phase = fmod(++phase_select,FsDivisor);
				// Restart the phase select counter if the phase is zero.
				if (!phase) phase_select = 0;
			}
		}
	}
	if (phase_select)
		return 1;
	else
		return 0;
}

// Write the contents of the data FIFO to the specified file.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long TargetSave(s_Adapter *p_HW, char *p_File, unsigned long ChannelNumber)
{
	unsigned long Status = 0;
	unsigned long long Value;
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	size_t data_items_count = 0;
	size_t record_count = 0;
	signed sample = 0;

	// Open the specified data file.
	fp = fopen(p_File, "w");

	// Process samples only if the file open was successful.
	if(fp != NULL)
	{
		// Sequence through every 64-bit record in the FIFO.
		for(record_count = 0; record_count < 2048; record_count++)
		{

			// Read the currrent FIFO value.
			Read(p_HW, BAR2, ChannelNumber << 12 | 0x030, &Value);
			// Pop the next FIFO value.
			Write(p_HW, BAR2, ChannelNumber << 12 | 0x030, 0x0ull);

			// Sequence through every sample in the 64-bit record.
			for (data_items_count = 0; data_items_count < 4; data_items_count++)
				{
					// Extract the next sample value from the 64-bit record.
					sample = (signed)((Value  >> (data_items_count * 16)) & (max >> 48));
					// Sign extend negative numbers.
					if (sample & (one << 11)) sample |= (max << 12);
					// Print the sample value to the file.
					fprintf(fp, "%d\n", sample);
				}
		}
		fclose(fp);
		Status = 0;
	}
	else
		Status = 1;

	return(Status);
}
